import 'dart:async';
import 'dart:ffi';
import 'dart:core';

import 'package:unit_rc/UnitRC.dart';

enum ProxyRCType {
  none,
  strong,
  weak,
}

abstract class DartBaseProxyClass {
  WeakPtr<T> weakBaseRC<T extends DartBaseClass>();

  T? strongBaseRC<T extends DartBaseClass>();
}

class WeakPtr<T extends DartBaseClass> {
  WeakReference<T>? ptr;

  WeakPtr({T? ptr}) {
    if (ptr != null) {
      this.ptr = WeakReference<T>(ptr);
    }
  }

  T? lock() {
    return ptr?.target?.strongRC();
  }
}

class DartBaseClass {
  final zone = Zone.current;

  Pointer<Void> strongPointer = nullptr;
  Pointer<Void> weakPointer = nullptr;

  Pointer<Void> toCpp({ProxyRCType rcType = ProxyRCType.strong}) {
    final isProxy = this.isProxy;
    if (strongPointer != nullptr && (isProxy || rcType == ProxyRCType.strong)) {
      return strongPointer;
    }
    if (weakPointer != nullptr && (isProxy || rcType == ProxyRCType.weak)) {
      return weakPointer;
    }
    return nullptr;
  }

  bool get isProxy {
    return this is DartBaseProxyClass;
  }

  void notifyFieldUpdated(int index) {
    _native_notifyFieldUpdated(toCpp(), index);
  }
  static late final _native_notifyFieldUpdated = unitRCLib.lookupFunction<
      Void Function(Pointer<Void>, Int32),
      void Function(
          Pointer<Void>, int)>("UnitRC_DartBaseClass_notifyFieldUpdated");
}

extension UnitRCWeakEx<T extends DartBaseClass> on T {
  T? strongRC() {
    return (this as DartBaseProxyClass?)?.strongBaseRC() ?? this;
  }

  WeakPtr<T> weakRC() {
    return (this as DartBaseProxyClass?)?.weakBaseRC() ?? WeakPtr(ptr: this);
  }
}